from abc import abstractmethod
from typing import Optional


class AudienceDeserializer:

    @abstractmethod
    def deserialize(self,
                    bytes_: bytes,
                    offset: int,
                    length: int) -> Optional[dict]:
        raise NotImplementedError



from abc import abstractmethod

from sdk.json.publish_event import PublishEvent


class ContextEventSerializer:

    @abstractmethod
    def serialize(self, publish_event: PublishEvent) -> bytearray:
        raise NotImplementedError



from abc import abstractmethod

from requests.adapters import Response


class HTTPClient:

    @abstractmethod
    def get(self,
            url: str,
            query: dict,
            headers: dict) -> Response:
        raise NotImplementedError

    @abstractmethod
    def put(self,
            url: str,
            query: dict,
            headers: dict,
            body: bytearray) -> Response:
        raise NotImplementedError

    @abstractmethod
    def post(self,
             url: str,
             query: dict,
             headers: dict,
             body: bytearray) -> Response:
        raise NotImplementedError



from typing import Optional

from sdk.context_event_logger import ContextEventLogger


class ContextConfig:
    refresh_interval: int = 50
    publish_delay: int = 50  # seconds
    event_logger: Optional[ContextEventLogger] = None
    cassigmnents: {} = None
    overrides: {} = None
    attributes: {} = None
    units: {} = None



from concurrent.futures import Future
from typing import Optional

from sdk.absmartly_config import ABSmartlyConfig
from sdk.audience_matcher import AudienceMatcher
from sdk.context import Context
from sdk.context_config import ContextConfig
from sdk.default_audience_deserializer import DefaultAudienceDeserializer
from sdk.default_context_data_provider import DefaultContextDataProvider
from sdk.default_context_event_handler import DefaultContextEventHandler
from sdk.default_variable_parser import DefaultVariableParser
from sdk.json.context_data import ContextData
from sdk.time.system_clock_utc import SystemClockUTC


class ABSmartly:

    def __init__(self, config: ABSmartlyConfig):
        self.context_data_provider = config.context_data_provider
        self.context_event_handler = config.context_event_handler
        self.context_event_logger = config.context_event_logger
        self.variable_parser = config.variable_parser
        self.audience_deserializer = config.audience_deserializer

        if self.context_data_provider is None or \
                self.context_event_handler is None:
            self.client = config.client

            if self.context_data_provider is None:
                self.context_data_provider = \
                    DefaultContextDataProvider(self.client)

            if self.context_event_handler is None:
                self.context_event_handler = \
                    DefaultContextEventHandler(self.client)

        if self.variable_parser is None:
            self.variable_parser = DefaultVariableParser()

        if self.audience_deserializer is None:
            self.audience_deserializer = DefaultAudienceDeserializer()

    def get_context_data(self) -> Future[Optional[ContextData]]:
        return self.context_data_provider.get_context_data()

    def create_context(self, config: ContextConfig) -> Context:
        return Context(SystemClockUTC(),
                       config,
                       self.context_data_provider.get_context_data(),
                       self.context_data_provider,
                       self.context_event_handler,
                       self.context_event_logger,
                       self.variable_parser,
                       AudienceMatcher(self.audience_deserializer))

    def create_context_with(self,
                            config: ContextConfig,
                            data: ContextData) -> Context:
        future_data = Future()
        future_data.set_result(data)
        return Context(SystemClockUTC(), config,
                       future_data,
                       self.context_data_provider,
                       self.context_event_handler,
                       self.context_event_logger,
                       self.variable_parser,
                       AudienceMatcher(self.audience_deserializer))



import requests as req
from requests.adapters import HTTPAdapter, Response
from urllib3 import Retry

from sdk.default_http_client_config import DefaultHTTPClientConfig
from sdk.http_client import HTTPClient


class DefaultHTTPClient(HTTPClient):

    def __init__(self, config: DefaultHTTPClientConfig):
        self.http_client = req.Session()
        retry = Retry(total=config.max_retries, read=config.max_retries,
                      connect=config.max_retries,
                      backoff_factor=config.retry_interval,
                      status_forcelist=(502, 503),
                      allowed_methods=frozenset(['GET', 'POST']))
        self.http_client.mount(
            'http://', HTTPAdapter(max_retries=retry,
                                   pool_maxsize=200,
                                   pool_connections=20))
        self.http_client.mount(
            'https://', HTTPAdapter(max_retries=retry,
                                    pool_maxsize=200,
                                    pool_connections=20))
        self.timeout = config.connection_timeout
        self.request_timeout = config.connection_request_timeout

    def get(self,
            url: str,
            query: dict,
            headers: dict) -> Response:
        return self.http_client.get(url,
                                    params=query,
                                    headers=headers,
                                    timeout=(self.timeout,
                                             self.request_timeout))

    def put(self,
            url: str,
            query: dict,
            headers: dict,
            body: bytearray) -> Response:
        headers.update({'Content-type': 'application/json'})
        return self.http_client.put(url,
                                    data=bytes(body),
                                    params=query,
                                    headers=headers,
                                    timeout=(self.timeout,
                                             self.request_timeout))

    def post(self,
             url: str,
             query: dict,
             headers: dict,
             body: bytearray) -> Response:
        headers.update({'Content-type': 'application/json'})
        return self.http_client.put(url,
                                    data=bytes(body),
                                    params=query,
                                    headers=headers,
                                    timeout=(self.timeout,
                                             self.request_timeout))



from abc import abstractmethod

from sdk.jsonexpr.evaluator import Evaluator


class Operator:

    @abstractmethod
    def evaluate(self, evaluator: Evaluator, args: object):
        raise NotImplementedError



from abc import abstractmethod

from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operator import Operator


class BooleanCombinator(Operator):

    def evaluate(self, evaluator: Evaluator, args: object):
        if type(args) is list:
            return self.combine(evaluator, args)
        else:
            return None

    @abstractmethod
    def combine(self, evaluator: Evaluator, args: list):
        raise NotImplementedError



import threading


class AtomicInt(object):
    def __init__(self):
        self.value = 0
        self._lock = threading.Lock()

    def set(self, value: int):
        with self._lock:
            self.value = value

    def increment(self):
        with self._lock:
            self.value += 1

    def increment_and_get(self):
        with self._lock:
            self.value += 1
            return self.value

    def get(self):
        with self._lock:
            return self.value



from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operators.boolean_combinator import BooleanCombinator


class OrCombinator(BooleanCombinator):
    def combine(self, evaluator: Evaluator, args: list):
        for item in args:
            if evaluator.boolean_convert(evaluator.evaluate(item)):
                return True
        return len(args) == 0



import unittest

from sdk.default_context_event_serializer import DefaultContextEventSerializer
from sdk.json.attribute import Attribute
from sdk.json.exposure import Exposure
from sdk.json.goal_achievement import GoalAchievement
from sdk.json.publish_event import PublishEvent
from sdk.json.unit import Unit


class ContextDataDeserializerTest(unittest.TestCase):

    def test_serialize(self):
        event = PublishEvent()
        event.publishedAt = 123456789
        event.hashed = True
        goal1 = GoalAchievement()
        goal1.name = "goal1"
        goal1.achievedAt = 123456000
        goal1.properties = {
            "amount": 6,
            "value": 5.0,
            "tries": 1,
            "nested": {"value": 5},
            "nested_arr": {"nested": [1, 2, "test"]}
        }

        goal2 = GoalAchievement()
        goal2.name = "goal2"
        goal2.achievedAt = 123456789
        goal2.properties = None
        event.goals = [goal1, goal2]

        unit = Unit()
        unit.type = "session_id"
        unit.uid = "pAE3a1i5Drs5mKRNq56adA"

        unit2 = Unit()
        unit2.type = "user_id"
        unit2.uid = "JfnnlDI7RTiF9RgfG2JNCw"

        event.units = [unit, unit2]

        exposure = Exposure()
        exposure.id = 1
        exposure.name = "exp_test_ab"
        exposure.unit = "session_id"
        exposure.variant = 1
        exposure.exposedAt = 123470000
        exposure.assigned = True
        exposure.eligible = True
        exposure.overridden = False
        exposure.fullOn = False
        exposure.custom = False
        exposure.audienceMismatch = True

        attribute = Attribute()
        attribute.name = "attr1"
        attribute.value = "value1"
        attribute.setAt = 123456000

        attribute1 = Attribute()
        attribute1.name = "attr2"
        attribute1.value = "value2"
        attribute1.setAt = 123456789

        attribute2 = Attribute()
        attribute2.name = "attr2"
        attribute2.value = None
        attribute2.setAt = 123450000

        attribute3 = Attribute()
        attribute3.name = "attr3"
        attribute3.value = {"nested": {"value": 5}}
        attribute3.setAt = 123470000

        attribute4 = Attribute()
        attribute4.name = "attr4"
        attribute4.value = {"nested": {"nested": [1, 2, "test"]}}
        attribute4.setAt = 123480000

        event.attributes = [attribute,
                            attribute1,
                            attribute2,
                            attribute3,
                            attribute4]

        serializer = DefaultContextEventSerializer()
        result = serializer.serialize(event)
        self.assertTrue("pAE3a1i5Drs5mKRNq56adA" in str(result, "utf-8"))



from concurrent.futures import Future
from typing import Optional

from sdk.client import Client
from sdk.context import Context
from sdk.context_event_handler import ContextEventHandler
from sdk.json.context_data import ContextData
from sdk.json.publish_event import PublishEvent


class DefaultContextEventHandler(ContextEventHandler):

    def __init__(self, client: Client):
        self.client = client

    def publish(self,
                context: Context,
                event: PublishEvent) -> Future[Optional[ContextData]]:
        return self.client.publish(event)



from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operator import Operator


class ValueOperator(Operator):
    def evaluate(self, evaluator: Evaluator, value: object):
        return value



import unittest
import sdk.internal.buffers as buffers
import sdk.internal.murmur32 as murmur


class BuffersTest(unittest.TestCase):
    def test_put_uint32(self):
        bytes_len = 9
        i = 0
        while i < bytes_len - 3:
            byte = bytearray(bytes_len)

            expected = (i + 1) * 0xe6546b64
            expected = murmur.to_signed32(expected)
            buffers.put_uint32(byte, i, expected)
            actual = murmur.to_signed32(buffers.get_uint32(byte, i))
            self.assertEqual(expected, actual)

            b = 0
            while b < bytes_len:
                if b < i or b >= (i + 4):
                    self.assertEqual(0, byte[b])
                b += 1

            i += 1

    def test_get_uint32(self):
        byte = bytearray([97, 226, 134, 147, 98, 196, 0])
        i = 0
        while i < len(byte) - 3:
            self.assertEqual((byte[i] & 0xff) |
                             ((byte[i + 1] & 0xff) << 8) |
                             ((byte[i + 2] & 0xff) << 16) | (
                        (byte[i + 3] & 0xff) << 24),
                             buffers.get_uint32(byte, i))
            i += 1

    def test_get_uint24(self):
        byte = bytearray([97, 226, 134, 147, 98, 0])
        i = 0
        while i < len(byte) - 2:
            self.assertEqual((byte[i] & 0xff) |
                             ((byte[i + 1] & 0xff) << 8) |
                             ((byte[i + 2] & 0xff) << 16),
                             buffers.get_uint24(byte, i))
            i += 1

    def test_get_uint16(self):
        byte = bytearray([97, 226, 134, 147, 98, 0])
        i = 0
        while i < len(byte) - 1:
            self.assertEqual((byte[i] & 0xff) |
                             ((byte[i + 1] & 0xff) << 8),
                             buffers.get_uint16(byte, i))
            i += 1

    def test_get_uint8(self):
        byte = bytearray([97, 226, 134, 147, 98, 0])
        i = 0
        while i < len(byte):
            self.assertEqual((byte[i] & 0xff), buffers.get_uint8(byte, i))
            i += 1



import threading


class AtomicBool(object):
    def __init__(self):
        self.value = False
        self._lock = threading.Lock()

    def set(self, value: bool):
        with self._lock:
            self.value = value

    def compare_and_set(self, expected_value: bool, new_value: bool):
        with self._lock:
            result = expected_value == self.value
            if result:
                self.value = new_value
            return result



from sdk.jsonexpr.expr_evaluator import ExprEvaluator
from sdk.jsonexpr.operators.and_combinator import AndCombinator
from sdk.jsonexpr.operators.equals_operator import EqualsOperator
from sdk.jsonexpr.operators.greater_than_operator import GreaterThanOperator
from sdk.jsonexpr.operators.greater_than_or_equal_operator \
    import GreaterThanOrEqualOperator
from sdk.jsonexpr.operators.in_operator import InOperator
from sdk.jsonexpr.operators.less_than_operator import LessThanOperator
from sdk.jsonexpr.operators.less_than_or_equal_operator \
    import LessThanOrEqualOperator
from sdk.jsonexpr.operators.match_operator import MatchOperator
from sdk.jsonexpr.operators.not_operator import NotOperator
from sdk.jsonexpr.operators.null_operator import NullOperator
from sdk.jsonexpr.operators.or_combinator import OrCombinator
from sdk.jsonexpr.operators.value_operator import ValueOperator
from sdk.jsonexpr.operators.var_operator import VarOperator


class JsonExpr:
    operators = {
        "and": AndCombinator(),
        "or": OrCombinator(),
        "value": ValueOperator(),
        "var": VarOperator(),
        "null": NullOperator(),
        "not": NotOperator(),
        "in": InOperator(),
        "match": MatchOperator(),
        "eq": EqualsOperator(),
        "gt": GreaterThanOperator(),
        "gte": GreaterThanOrEqualOperator(),
        "lt": LessThanOperator(),
        "lte": LessThanOrEqualOperator()
    }

    def evaluate_boolean_expr(self, expr: object, var: dict):
        evaluator = ExprEvaluator(self.operators, var)
        return evaluator.boolean_convert(evaluator.evaluate(expr))

    def evaluate_expr(self, expr: object, var: dict):
        evaluator = ExprEvaluator(self.operators, var)
        return evaluator.evaluate(expr)



from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operators.unary_operator import UnaryOperator


class NullOperator(UnaryOperator):
    def unary(self, evaluator: Evaluator, arg: object):
        return arg is None



from abc import abstractmethod


class Evaluator:

    @abstractmethod
    def evaluate(self, expr: object):
        raise NotImplementedError

    @abstractmethod
    def boolean_convert(self, x: object):
        raise NotImplementedError

    @abstractmethod
    def number_convert(self, x: object):
        raise NotImplementedError

    @abstractmethod
    def string_convert(self, x: object):
        raise NotImplementedError

    @abstractmethod
    def extract_var(self, path: str):
        raise NotImplementedError

    # returns
    # -1 -> lesser, 0 -> equals, 1 -> greater, null -> undefined comparison
    @abstractmethod
    def compare(self, lhs: object, rhs: object):
        raise NotImplementedError



import base64
import hashlib
import unittest

import sdk.internal.variant_assigner as assigner


class VariantAssignerTest(unittest.TestCase):
    def test_choose_variant(self):
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.0, 1.0], 0.0))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.0, 1.0], 0.5))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.0, 1.0], 1.0))

        self.assertEqual(0, assigner.VariantAssigner.choose_variant(
            [1.0, 0.0], 0.0))
        self.assertEqual(0, assigner.VariantAssigner.choose_variant(
            [1.0, 0.0], 0.5))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [1.0, 0.0], 1.0))

        self.assertEqual(0, assigner.VariantAssigner.choose_variant(
            [0.5, 0.5], 0.0))
        self.assertEqual(0, assigner.VariantAssigner.choose_variant(
            [0.5, 0.5], 0.25))
        self.assertEqual(0, assigner.VariantAssigner.choose_variant(
            [0.5, 0.5], 0.49999999))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.5, 0.5], 0.5))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.5, 0.5], 0.50000001))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.5, 0.5], 0.75))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.5, 0.5], 1.0))

        self.assertEqual(0, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.0))
        self.assertEqual(0, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.25))
        self.assertEqual(0, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.33299999))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.333))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.33300001))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.5))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.66599999))
        self.assertEqual(2, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.666))
        self.assertEqual(2, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.66600001))
        self.assertEqual(2, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 0.75))
        self.assertEqual(2, assigner.VariantAssigner.choose_variant(
            [0.333, 0.333, 0.334], 1.0))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.0, 1.0], 0.0))
        self.assertEqual(1, assigner.VariantAssigner.choose_variant(
            [0.0, 1.0], 1.0))

    def test_assignments_match(self):
        splits = [[0.5, 0.5],
                  [0.5, 0.5],
                  [0.5, 0.5],
                  [0.5, 0.5],
                  [0.5, 0.5],
                  [0.5, 0.5],
                  [0.5, 0.5],
                  [0.33, 0.33, 0.34],
                  [0.33, 0.33, 0.34],
                  [0.33, 0.33, 0.34],
                  [0.33, 0.33, 0.34],
                  [0.33, 0.33, 0.34],
                  [0.33, 0.33, 0.34],
                  [0.33, 0.33, 0.34]]

        seeds = [[0x00000000, 0x00000000],
                 [0x00000000, 0x00000001],
                 [0x8015406f, 0x7ef49b98],
                 [0x3b2e7d90, 0xca87df4d],
                 [0x52c1f657, 0xd248bb2e],
                 [0x865a84d0, 0xaa22d41a],
                 [0x27d1dc86, 0x845461b9],
                 [0x00000000, 0x00000000],
                 [0x00000000, 0x00000001],
                 [0x8015406f, 0x7ef49b98],
                 [0x3b2e7d90, 0xca87df4d],
                 [0x52c1f657, 0xd248bb2e],
                 [0x865a84d0, 0xaa22d41a],
                 [0x27d1dc86, 0x845461b9]]

        unituid = "bleh@absmartly.com"
        dig = hashlib.md5(unituid.encode('utf-8')).digest()
        unithash = base64.urlsafe_b64encode(dig).rstrip(b'=')
        var_assigner = assigner.VariantAssigner(bytearray(unithash))
        expected_variants = [0, 1, 0, 0, 0, 0, 1, 0, 2, 0, 0, 0, 1, 1]

        self.assert_variants(seeds, splits, expected_variants, var_assigner)

        unituid = str(123456789)
        dig = hashlib.md5(unituid.encode('utf-8')).digest()
        unithash = base64.urlsafe_b64encode(dig).rstrip(b'=')
        var_assigner = assigner.VariantAssigner(bytearray(unithash))
        expected_variants = [1, 0, 1, 1, 1, 0, 0, 2, 1, 2, 2, 2, 0, 0]

        self.assert_variants(seeds, splits, expected_variants, var_assigner)

        unituid = "e791e240fcd3df7d238cfc285f475e8152fcc0ec"
        dig = hashlib.md5(unituid.encode('utf-8')).digest()
        unithash = base64.urlsafe_b64encode(dig).rstrip(b'=')
        var_assigner = assigner.VariantAssigner(bytearray(unithash))
        expected_variants = [1, 0, 1, 1, 0, 0, 0, 2, 0, 2, 1, 0, 0, 1]

        self.assert_variants(seeds, splits, expected_variants, var_assigner)

    def assert_variants(self, seeds, splits, expected_variants, var_assigner):
        for index, seed in enumerate(seeds):
            frags = seed
            split = splits[index]
            variant = var_assigner.assign(split, frags[0], frags[1])
            self.assertEqual(expected_variants[index], variant)



import typing

from sdk.json.attribute import Attribute
from sdk.json.exposure import Exposure
from sdk.json.goal_achievement import GoalAchievement
from sdk.json.unit import Unit


class PublishEvent:
    hashed: typing.Optional[bool] = False
    units: typing.Optional[list[Unit]] = None
    publishedAt: typing.Optional[int] = 0
    exposures: typing.Optional[list[Exposure]] = None
    goals: typing.Optional[list[GoalAchievement]] = None
    attributes: typing.Optional[list[Attribute]] = None



import time
import unittest
from unittest.mock import MagicMock, Mock

from requests import Response

from sdk.client import Client
from sdk.client_config import ClientConfig
from sdk.default_context_data_deserializer import \
    DefaultContextDataDeserializer
from sdk.default_context_event_serializer import \
    DefaultContextEventSerializer
from sdk.default_http_client import DefaultHTTPClient
from sdk.default_http_client_config import DefaultHTTPClientConfig
from sdk.json.context_data import ContextData
from sdk.json.publish_event import PublishEvent


class ClientTest(unittest.TestCase):

    def test_create_with_defaults(self):
        config = ClientConfig()
        config.endpoint = "https://localhost/v1"
        config.api_key = "test-api-key"
        config.application = "website"
        config.environment = "dev"

        expected = ContextData()
        event = PublishEvent()
        publish_bytes = bytes(0)

        expected_query = {"application": "website", "environment": "dev"}

        deserializer = DefaultContextDataDeserializer()
        deserializer.deserialize = MagicMock(return_value=expected)

        serializer = DefaultContextEventSerializer()
        serializer.serialize = MagicMock(return_value=publish_bytes)
        http_client = DefaultHTTPClient(DefaultHTTPClientConfig())
        http_client.get = MagicMock(return_value="test")
        http_client.put = Mock(return_value="test")

        client = Client(config, http_client)
        client.get_context_data()

        http_client.get.assert_called_once_with(
            "https://localhost/v1/context",
            expected_query,
            {})
        http_client.get.reset_mock()
        client.publish(event)
        time.sleep(0.1)
        http_client.put.assert_called()

    def test_get_context_data(self):
        config = ClientConfig()
        config.endpoint = "https://localhost/v1"
        config.api_key = "test-api-key"
        config.application = "website"
        config.environment = "dev"

        expected = ContextData()
        publish_bytes = bytes(0)
        expected_query = {"application": "website", "environment": "dev"}

        deserializer = DefaultContextDataDeserializer()
        deserializer.deserialize = MagicMock(return_value=expected)

        serializer = DefaultContextEventSerializer()
        serializer.serialize = MagicMock(return_value=publish_bytes)
        http_client = DefaultHTTPClient(DefaultHTTPClientConfig())

        response = Response()
        response.status_code = 200
        response._content = bytes("{}", encoding="utf-8")
        http_client.get = MagicMock(return_value=response)
        http_client.put = Mock(return_value="test")

        client = Client(config, http_client)
        future = client.get_context_data()

        http_client.get.assert_called_once_with(
            "https://localhost/v1/context",
            expected_query,
            {})
        http_client.get.reset_mock()

        result = future.result()
        self.assertEqual(type(expected), type(result))

    def test_publish(self):
        config = ClientConfig()
        config.endpoint = "https://localhost/v1"
        config.api_key = "test-api-key"
        config.application = "website"
        config.environment = "dev"

        expected = ContextData()

        event = PublishEvent()
        publish_bytes = bytes(0)

        expected_headers = {"X-API-Key": "test-api-key",
                            "X-Application": "website",
                            "X-Environment": "dev",
                            "X-Application-Version": '0',
                            "X-Agent": "absmartly-python-sdk"}

        deserializer = DefaultContextDataDeserializer()
        deserializer.deserialize = MagicMock(return_value=expected)

        serializer = DefaultContextEventSerializer()
        serializer.serialize = MagicMock(return_value=publish_bytes)
        http_client = DefaultHTTPClient(DefaultHTTPClientConfig())

        response = Response()
        response.status_code = 200
        response._content = bytes("{}", encoding="utf-8")
        http_client.put = Mock(return_value=response)

        client = Client(config, http_client)
        client.serializer = serializer
        client.publish(event)

        http_client.put.assert_called_once_with("https://localhost/v1/context",
                                                {},
                                                expected_headers,
                                                publish_bytes)
        http_client.put.reset_mock()
        serializer.serialize.assert_called_once()



from setuptools import setup

setup(
    name='absmartly',
    version='0.1.0',
    description='A Python SDK lib',
    url='https://github.com/absmartly/python-sdk',
    author='ABSmartly',
    author_email='pram@gmail.com',
    license='BSD 2-clause',
    packages=['sdk'],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



from typing import Optional

import jsons
from jsons import DeserializationError

from sdk.context_data_deserializer import ContextDataDeserializer
from sdk.json.context_data import ContextData


class DefaultContextDataDeserializer(ContextDataDeserializer):
    def deserialize(self,
                    bytes_: bytes,
                    offset: int,
                    length: int) -> Optional[ContextData]:
        try:
            return jsons.loadb(bytes_, ContextData)
        except DeserializationError:
            return None



import typing


class Exposure:
    id: typing.Optional[int] = 0
    name: typing.Optional[str] = None
    unit: typing.Optional[str] = None
    variant: typing.Optional[int] = 0
    exposedAt: typing.Optional[int] = 0
    assigned: typing.Optional[bool] = False
    eligible: typing.Optional[bool] = False
    overridden: typing.Optional[bool] = False
    fullOn: typing.Optional[bool] = False
    custom: typing.Optional[bool] = False
    audienceMismatch: typing.Optional[bool] = False



from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operators.binary_operator import BinaryOperator


class EqualsOperator(BinaryOperator):
    def binary(self, evaluator: Evaluator, lhs: object, rhs: object):
        result = evaluator.compare(lhs, rhs)
        if result is not None:
            return result == 0
        else:
            return None



from abc import abstractmethod

from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operator import Operator


class BinaryOperator(Operator):
    def evaluate(self, evaluator: Evaluator, args: object):
        if type(args) is list:
            lhs = evaluator.evaluate(args[0]) if len(args) > 0 else None
            if lhs is not None:
                rhs = evaluator.evaluate(args[1]) if len(args) > 1 else None
                if rhs is not None:
                    return self.binary(evaluator, lhs, rhs)
        return None

    @abstractmethod
    def binary(self, evaluator: Evaluator, lhs: object, rhs: object):
        raise NotImplementedError



from typing import Optional

from sdk.audience_deserializer import AudienceDeserializer
from sdk.jsonexpr.json_expr import JsonExpr


class Result:
    def __init__(self, result: bool):
        self.result = result


class AudienceMatcher:

    def __init__(self, deserializer: AudienceDeserializer):
        self.deserializer = deserializer
        self.json_expr = JsonExpr()

    def evaluate(self, audience: str, attributes: dict) -> Optional[Result]:
        bytes_arr = bytes(audience, encoding="utf-8")
        audience_map = self.deserializer.deserialize(bytes_arr,
                                                     0,
                                                     len(bytes_arr))
        if audience_map is not None:
            if "filter" in audience_map:
                fl = audience_map["filter"]
                if type(fl) is dict or type(fl) is list:
                    expr = self.json_expr.evaluate_boolean_expr(fl,
                                                                attributes)
                    return Result(expr)

        return None



from abc import abstractmethod
from enum import Enum


class EventType(Enum):
    ERROR = "error"
    READY = "ready"
    REFRESH = "refresh"
    PUBLISH = "publish"
    EXPOSURE = "exposure"
    GOAL = "goal"
    CLOSE = "close"


class ContextEventLogger:

    @abstractmethod
    def handle_event(self, event_type: EventType, data: object):
        raise NotImplementedError
